@@ -19,13 +19,16 @@ import threading
 
 from concurrent import futures as _futures
 from concurrent.futures import process as _process
-from concurrent.futures import thread as _thread
 import six
 
+from six.moves import queue as compat_queue
+
 from futurist import _green
+from futurist import _thread
 from futurist import _utils
 
 
+TimeoutError = _futures.TimeoutError
 CancelledError = _futures.CancelledError
 
 
@@ -37,25 +40,6 @@ class RejectedSubmission(Exception):
 Future = _futures.Future
 
 
-class _Threading(object):
-
-    @staticmethod
-    def event_object(*args, **kwargs):
-        return threading.Event(*args, **kwargs)
-
-    @staticmethod
-    def lock_object(*args, **kwargs):
-        return threading.Lock(*args, **kwargs)
-
-    @staticmethod
-    def rlock_object(*args, **kwargs):
-        return threading.RLock(*args, **kwargs)
-
-    @staticmethod
-    def condition_object(*args, **kwargs):
-        return threading.Condition(*args, **kwargs)
-
-
 class _Gatherer(object):
     def __init__(self, submit_func, lock_factory, start_before_submit=False):
         self._submit_func = submit_func
@@ -116,7 +100,7 @@ class _Gatherer(object):
         return fut
 
 
-class ThreadPoolExecutor(_thread.ThreadPoolExecutor):
+class ThreadPoolExecutor(_futures.Executor):
     """Executor that uses a thread pool to execute calls asynchronously.
 
     It gathers statistics about the submissions executed for post-analysis...
@@ -124,13 +108,13 @@ class ThreadPoolExecutor(_thread.ThreadPoolExecutor):
     See: https://docs.python.org/dev/library/concurrent.futures.html
     """
 
-    threading = _Threading()
+    threading = _thread.Threading()
 
     def __init__(self, max_workers=None, check_and_reject=None):
         """Initializes a thread pool executor.
 
         :param max_workers: maximum number of workers that can be
-                            simulatenously active at the same time, further
+                            simultaneously active at the same time, further
                             submitted work will be queued up when this limit
                             is reached.
         :type max_workers: int
@@ -146,21 +130,15 @@ class ThreadPoolExecutor(_thread.ThreadPoolExecutor):
         """
         if max_workers is None:
             max_workers = _utils.get_optimal_thread_count()
-        super(ThreadPoolExecutor, self).__init__(max_workers=max_workers)
-        if self._max_workers <= 0:
+        if max_workers <= 0:
             raise ValueError("Max workers must be greater than zero")
-        # NOTE(harlowja): this replaces the parent classes non-reentrant lock
-        # with a reentrant lock so that we can correctly call into the check
-        # and reject lock, and that it will block correctly if another
-        # submit call is done during that...
+        self._max_workers = max_workers
+        self._work_queue = compat_queue.Queue()
         self._shutdown_lock = threading.RLock()
+        self._shutdown = False
+        self._workers = []
         self._check_and_reject = check_and_reject or (lambda e, waiting: None)
-        self._gatherer = _Gatherer(
-            # Since our submit will use this gatherer we have to reference
-            # the parent submit, bound to this instance (which is what we
-            # really want to use anyway).
-            super(ThreadPoolExecutor, self).submit,
-            self.threading.lock_object)
+        self._gatherer = _Gatherer(self._submit, self.threading.lock_object)
 
     @property
     def statistics(self):
@@ -172,6 +150,35 @@ class ThreadPoolExecutor(_thread.ThreadPoolExecutor):
         """Accessor to determine if the executor is alive/active."""
         return not self._shutdown
 
+    def _maybe_spin_up(self):
+        """Spin up a worker if needed."""
+        # Do more advanced idle checks and/or reaping of very idle
+        # threads in the future....
+        if (not self._workers or
+                len(self._workers) < self._max_workers):
+            w = _thread.ThreadWorker.create_and_register(
+                self, self._work_queue)
+            # Always save it before we start (so that even if we fail
+            # starting it we can correctly join on it).
+            self._workers.append(w)
+            w.start()
+
+    def shutdown(self, wait=True):
+        with self._shutdown_lock:
+            if not self._shutdown:
+                self._shutdown = True
+                for w in self._workers:
+                    w.stop()
+        if wait:
+            for w in self._workers:
+                _thread.join_thread(w)
+
+    def _submit(self, fn, *args, **kwargs):
+        f = Future()
+        self._maybe_spin_up()
+        self._work_queue.put(_utils.WorkItem(f, fn, args, kwargs))
+        return f
+
     def submit(self, fn, *args, **kwargs):
         """Submit some work to be executed (and gather statistics)."""
         with self._shutdown_lock:
@@ -190,11 +197,11 @@ class ProcessPoolExecutor(_process.ProcessPoolExecutor):
     See: https://docs.python.org/dev/library/concurrent.futures.html
     """
 
-    threading = _Threading()
+    threading = _thread.Threading()
 
     def __init__(self, max_workers=None):
         if max_workers is None:
-            max_workers = _utils.get_optimal_thread_count()
+            max_workers = _utils.get_optimal_process_count()
         super(ProcessPoolExecutor, self).__init__(max_workers=max_workers)
         if self._max_workers <= 0:
             raise ValueError("Max workers must be greater than zero")
@@ -231,7 +238,7 @@ class SynchronousExecutor(_futures.Executor):
     It gathers statistics about the submissions executed for post-analysis...
     """
 
-    threading = _Threading()
+    threading = _thread.Threading()
 
     def __init__(self, green=False, run_work_func=lambda work: work.run()):
         """Synchronous executor constructor.
@@ -308,7 +315,7 @@ class GreenFuture(Future):
         # functions will correctly yield to eventlet. If this is not done then
         # waiting on the future never actually causes the greenthreads to run
         # and thus you wait for infinity.
-        if not _green.is_monkey_patched('threading'):
+        if not _green.is_monkey_patched('thread'):
             self._condition = _green.threading.condition_object()
 
 
@@ -408,8 +415,8 @@ class GreenThreadPoolExecutor(_futures.Executor):
             else:
                 shutoff = False
         if wait and shutoff:
-            self._pool.waitall()
             self._delayed_work.join()
+            self._pool.waitall()
 
 
 class ExecutorStatistics(object):
