@@ -16,10 +16,10 @@ import contextlib
 import functools
 import threading
 import time
+from unittest import mock
 
 import eventlet
 from eventlet.green import threading as green_threading
-import mock
 import testscenarios
 
 import futurist
@@ -117,6 +117,38 @@ class TestPeriodics(testscenarios.TestWithScenarios, base.TestCase):
         schedule_order = w._schedule._ordering
         self.assertEqual([(expected_next, 0)], schedule_order)
 
+    def _run_work_up_to(self, stop_after_num_calls):
+        ev = self.event_cls()
+        called_tracker = []
+
+        @periodics.periodic(0.1, run_immediately=True)
+        def fast_periodic():
+            called_tracker.append(True)
+            if len(called_tracker) >= stop_after_num_calls:
+                ev.set()
+
+        callables = [
+            (fast_periodic, None, None),
+        ]
+
+        worker_kwargs = self.worker_kwargs.copy()
+        w = periodics.PeriodicWorker(callables, **worker_kwargs)
+
+        with self.create_destroy(w.start):
+            ev.wait()
+            w.stop()
+
+        return list(w.iter_watchers())[0], fast_periodic
+
+    def test_watching_work(self):
+        for i in [3, 5, 9, 11]:
+            watcher, cb = self._run_work_up_to(i)
+            self.assertEqual(cb, watcher.work.callback)
+            self.assertGreaterEqual(i, watcher.runs)
+            self.assertGreaterEqual(i, watcher.successes)
+            self.assertEqual((), watcher.work.args)
+            self.assertEqual({}, watcher.work.kwargs)
+
     def test_last_finished_strategy(self):
         last_now = 3.2
         nows = [
@@ -134,6 +166,30 @@ class TestPeriodics(testscenarios.TestWithScenarios, base.TestCase):
         nows = list(reversed(nows))
         self._test_strategy('last_finished', nows, last_now, 5.0)
 
+    def test_last_finished_strategy_jitter(self):
+        ev = self.event_cls()
+        calls = []
+        stop_after_calls = 3
+
+        @periodics.periodic(0.1, run_immediately=False)
+        def fast_periodic():
+            calls.append(True)
+            if len(calls) > stop_after_calls:
+                ev.set()
+
+        worker_kwargs = self.worker_kwargs.copy()
+        worker_kwargs['schedule_strategy'] = 'last_finished_jitter'
+        callables = [
+            (fast_periodic, None, None),
+        ]
+        w = periodics.PeriodicWorker(callables, **worker_kwargs)
+
+        with self.create_destroy(w.start):
+            ev.wait()
+            w.stop()
+
+        self.assertGreaterEqual(len(calls), stop_after_calls)
+
     def test_waiting_immediate_add_processed(self):
         ran_at = []
 
@@ -176,6 +232,38 @@ class TestPeriodics(testscenarios.TestWithScenarios, base.TestCase):
         nows = list(reversed(nows))
         self._test_strategy('last_started', nows, last_now, 4.0)
 
+    def test_failure_callback_fail(self):
+        worker_kwargs = self.worker_kwargs.copy()
+        worker_kwargs['on_failure'] = 'not-a-func'
+        self.assertRaises(ValueError,
+                          periodics.PeriodicWorker, [], **worker_kwargs)
+
+    def test_failure_callback(self):
+        captures = []
+        ev = self.event_cls()
+
+        @periodics.periodic(0.1)
+        def failing_always():
+            raise RuntimeError("Broke!")
+
+        def trap_failures(cb, kind, periodic_spacing,
+                          exc_info, traceback=None):
+            captures.append([cb, kind, periodic_spacing, traceback])
+            ev.set()
+
+        callables = [
+            (failing_always, None, None),
+        ]
+        worker_kwargs = self.worker_kwargs.copy()
+        worker_kwargs['on_failure'] = trap_failures
+        w = periodics.PeriodicWorker(callables, **worker_kwargs)
+        with self.create_destroy(w.start, allow_empty=True):
+            ev.wait()
+            w.stop()
+
+        self.assertEqual(captures[0], [failing_always, periodics.PERIODIC,
+                                       0.1, mock.ANY])
+
     def test_aligned_strategy(self):
         last_now = 5.5
         nows = [
@@ -235,6 +323,19 @@ class TestPeriodics(testscenarios.TestWithScenarios, base.TestCase):
         self.assertIsNotNone(w.add(add_me))
         self.assertEqual(1, len(w))
 
+    def test_interval_checking(self):
+
+        @periodics.periodic(-0.5, enabled=False)
+        def no_add_me():
+            pass
+
+        w = periodics.PeriodicWorker([], **self.worker_kwargs)
+        self.assertEqual(0, len(w))
+        self.assertIsNone(w.add(no_add_me))
+        self.assertEqual(0, len(w))
+
+        self.assertRaises(ValueError, periodics.periodic, -0.5)
+
     def test_is_periodic(self):
 
         @periodics.periodic(0.5, enabled=False)
@@ -346,6 +447,83 @@ class TestPeriodics(testscenarios.TestWithScenarios, base.TestCase):
 
         m.assert_called_with('foo', bar='baz')
 
+    def test_never_again_exc(self):
+
+        m_1 = mock.MagicMock()
+        m_2 = mock.MagicMock()
+
+        @periodics.periodic(0.5)
+        def run_only_once():
+            m_1()
+            raise periodics.NeverAgain("No need to run again !!")
+
+        @periodics.periodic(0.5)
+        def keep_running():
+            m_2()
+
+        callables = [
+            (run_only_once, None, None),
+            (keep_running, None, None),
+        ]
+        executor_factory = lambda: self.executor_cls(**self.executor_kwargs)
+        w = periodics.PeriodicWorker(callables,
+                                     executor_factory=executor_factory,
+                                     **self.worker_kwargs)
+        with self.create_destroy(w.start):
+            self.sleep(2.0)
+            w.stop()
+
+        for watcher in w.iter_watchers():
+            self.assertGreaterEqual(watcher.runs, 1)
+            self.assertGreaterEqual(watcher.successes, 1)
+            self.assertEqual(watcher.failures, 0)
+
+        self.assertEqual(m_1.call_count, 1)
+        self.assertGreaterEqual(m_2.call_count, 3)
+
+    def test_start_with_auto_stop_when_empty_set(self):
+
+        @periodics.periodic(0.5)
+        def run_only_once():
+            raise periodics.NeverAgain("No need to run again !!")
+
+        callables = [
+            (run_only_once, None, None),
+            (run_only_once, None, None),
+        ]
+        executor_factory = lambda: self.executor_cls(**self.executor_kwargs)
+        w = periodics.PeriodicWorker(callables,
+                                     executor_factory=executor_factory,
+                                     **self.worker_kwargs)
+        with self.create_destroy(w.start, auto_stop_when_empty=True):
+            self.sleep(2.0)
+
+        for watcher in w.iter_watchers():
+            self.assertGreaterEqual(watcher.runs, 1)
+            self.assertGreaterEqual(watcher.successes, 1)
+            self.assertEqual(watcher.failures, 0)
+            self.assertEqual(watcher.requested_stop, True)
+
+    def test_add_with_auto_stop_when_empty_set(self):
+        m = mock.Mock()
+
+        @periodics.periodic(0.5)
+        def run_only_once():
+            raise periodics.NeverAgain("No need to run again !!")
+
+        callables = [
+            (run_only_once, None, None),
+        ]
+        executor_factory = lambda: self.executor_cls(**self.executor_kwargs)
+        w = periodics.PeriodicWorker(callables,
+                                     executor_factory=executor_factory,
+                                     **self.worker_kwargs)
+        with self.create_destroy(w.start, auto_stop_when_empty=True):
+            self.sleep(2.0)
+            w.add(every_half_sec, m, None)
+
+        m.assert_not_called()
+
 
 class RejectingExecutor(futurist.GreenThreadPoolExecutor):
     MAX_REJECTIONS_COUNT = 2
@@ -360,6 +538,32 @@ class RejectingExecutor(futurist.GreenThreadPoolExecutor):
         super(RejectingExecutor, self).__init__(check_and_reject=self._reject)
 
 
+class TestPformat(base.TestCase):
+    def test_invalid(self):
+
+        @periodics.periodic(1)
+        def a():
+            pass
+
+        @periodics.periodic(2)
+        def b():
+            pass
+
+        @periodics.periodic(3)
+        def c():
+            pass
+
+        callables = [
+            (a, None, None),
+            (b, None, None),
+            (c, None, None)
+        ]
+        w = periodics.PeriodicWorker(callables)
+
+        self.assertRaises(ValueError, w.pformat, columns=[])
+        self.assertRaises(ValueError, w.pformat, columns=['not a column'])
+
+
 class TestRetrySubmission(base.TestCase):
     def test_retry_submission(self):
         called = []
